/*
 * File: Yahtzee.java
 * ------------------
 * This program will eventually play the Yahtzee game.
 */

import acm.io.*;
import acm.program.*;
import acm.util.*;
import java.io.*;

public class Yahtzee extends GraphicsProgram implements YahtzeeConstants {
	
	private static final int playerTimesNumber = 3;
	private static final int CHEATMODE = -99;
	private static final int UPPERBONUS = 35;
	private static final int UPPERVALUE = 63;
	private static final int YAHTZEE_BONUS_CLIP = 100;
	
	// really need to check the number of players is not above 
	//MAX_PLAYERS
	public void run() {
		IODialog dialog = getDialog();
		nPlayers = dialog.readInt("Enter number of players");
		if (nPlayers == CHEATMODE) 
		{
			cheatMode = true;
			nPlayers=1;
		}
		nXPlayer =4;
		while (nXPlayer >= 4)
		{
			nXPlayer = dialog.readInt("Enter X player boards (3 or lower and higher than 0)");
		}
		playerNames = new String[nPlayers*nXPlayer];
		for (int i = 1; i <= nPlayers; i++) {
			String playerName = dialog.readLine("Enter name for player "+i);
			for (int j =0; j < nXPlayer; j++)
			{
				playerNames[(i-1)+(j+(i-1))] = playerName + " : "+j;
			}
		}
		display = new YahtzeeDisplay(getGCanvas(), playerNames);
		playGame();
	}

	/* setup the dice array, the dice is the dice to setup and the 
	 * second bool setup parameter is to either setup the whole (start) or just some
	 */
	private void setupDice(int[] dice, Boolean setup)
	{
		for (int i =0; i < YahtzeeConstants.N_DICE; i++)
		{
			if (setup) 
				dice[i] = rgen.nextInt(1,6);
			else if (display.isDieSelected(i))
				dice[i] = rgen.nextInt(1, 6);
		}
	}

	// false means that it has been updated, else true if there is a value already
	private Boolean updateScore(int[][] scoresUpdate, int category, int score, int boardX)
	{
		// default is -1, since cannot get this score value
		if (scoresUpdate[boardX][category] == -1)
		{
			scoresUpdate[boardX][category] = score*(boardX+1);
			return false;
		}
		// add in the bonus clip for multiple yahtzee only if over 0 already
		else if (category == YahtzeeConstants.YAHTZEE && scoresUpdate[boardX][category] > 0 && score > 0)
		{
			scoresUpdate[boardX][category] += (YAHTZEE_BONUS_CLIP*(boardX+1));
			score = scoresUpdate[boardX][category];
			return false;
		} else
		{
			display.printMessage("Category already selected, please select another");
			return true;
		}
	}
	
	/* setup the scores to a value of -1, default value */
	private void setupScores(int[][][] scoresSetup, int playersNum, int totalValues, int playerXBoards)
	{
		for (int i=0; i < playersNum; i++)
			for (int k =0; k < playerXBoards; k++)
				for (int j=0; j < totalValues;j++)
					scoresSetup[i][k][j] = -1;
	}

	// add the total scores together
	private int totalScores(int[] scores)
	{
		int returnValue = 0;
		for (int i=0; i < N_CATEGORIES; i++)
		{
			returnValue += (scores[i] == -1 ? 0 : scores[i]);
		}
		return returnValue;
	}

	// to find the values, first order the dice, thus able to create the straight and also if the dice are
	// sorted then can easily find the ones, twos etc.
	private void orderDice(int[] dice)
	{
		// use the bubble sort
		for (int i =0; i < YahtzeeConstants.N_DICE; i++)
		{
			for (int j =0; j < YahtzeeConstants.N_DICE; j++)
			{
				// swap if right value is greater than the left
				if (dice[i] < dice[j])
				{
					int temp = dice[i];
					dice[i] = dice[j];
					dice[j] = temp; 
				}
			}
		}
	}
	
	// X being the three / four / five (Yahtzee!!)
	private int checkXOfAKind(int[] dice, int X)
	{
		// if count = 3 then there is a three of a kind 
		int count =1, sum = dice[0];
		for (int i =1; i < YahtzeeConstants.N_DICE; i++)
		{
			// if the same value
			if (dice[i] == dice[i-1])
			{
				count++;
			}
			else
			{
				if (count < X) 
					count = 1; 
			}
			sum += dice[i];
		}
		if (count >= X) 
			return sum;
		else 
			return 0;
	}
	
	// need to check for 3 of the same value and then 2 of the same value
	private Boolean checkFullHouse(int[] dice)
	{
		int count =1;
		Boolean found3 = false, found2 = false;
		for (int i=1; i < YahtzeeConstants.N_DICE;i++)
		{
			if (dice[i] == dice[i-1]) 
				count++;
			else
			{
				if (count == 3)
					found3 = true;
				else if (count == 2)
					found2 = true;
				count=1;
			}
		}
		if (count == 3)
			found3 = true;
		else if (count == 2)
			found2 = true;
		if (found3 && found2) return true;
		return false;
	}

	// add single categories e.g. 1,2,3,4,5 and only the dice that have those values
	private int addSingle(int[] dice, int category)
	{
		int sum =0;
		for (int i =0; i < YahtzeeConstants.N_DICE; i++) 
		{
			if (category == dice[i]) sum+=dice[i];
			if (category == YahtzeeConstants.CHANCE)sum+=dice[i]; 
		}
		return sum;
	}
	
	// consecutiveValues = how many in a straight line
	private Boolean checkStraight(int[] dice, int consecutiveValues)
	{
		int inLine =1;
		for (int i =1; i < YahtzeeConstants.N_DICE; i++)
		{
			if ((dice[i-1]+1) == dice[i]) inLine++;
		}
		if (inLine >= consecutiveValues) return true;
		return false;
	}
	
	// update the scores after the game, e.g. upper/lower values and also add in the upper score
	private int updateScoreAfterGame(int[] scores)
	{
		int sum=0, returnTotal=0;
		for (int i =1; i < YahtzeeConstants.UPPER_SCORE; i++)
		{
			sum+= scores[i];
		}
		if (sum >= UPPERVALUE) 
		{
			scores[YahtzeeConstants.UPPER_BONUS-1] = UPPERBONUS;
			returnTotal = UPPERBONUS;
		}
		else
			scores[YahtzeeConstants.UPPER_BONUS-1] = 0;
		scores[YahtzeeConstants.UPPER_SCORE-1] = sum;
		returnTotal += sum;
		sum =0;
		for (int i =YahtzeeConstants.UPPER_SCORE+2; i < YahtzeeConstants.LOWER_SCORE; i++)
		{
			sum += scores[i];
		}
		scores[YahtzeeConstants.LOWER_SCORE-1] = sum;
		returnTotal += sum;
		return returnTotal;
	}
	
	private int checkCategory(int[] dice, int category)
	{
		int returnValue=0;
		Boolean straight =false;
		switch (category)
		{
			// all of these are the same, because it does not matter what values are inside the dice array
			case YahtzeeConstants.ONES :  
			case YahtzeeConstants.TWOS :  
			case YahtzeeConstants.THREES :
			case YahtzeeConstants.FOURS : 
			case YahtzeeConstants.FIVES : 
			case YahtzeeConstants.SIXES :
			case YahtzeeConstants.CHANCE :
					orderDice(dice);
					returnValue = addSingle(dice, category);
					break;
			case YahtzeeConstants.THREE_OF_A_KIND :
					// need to check for a three of a kind e.g.
					orderDice(dice);
					returnValue =  checkXOfAKind(dice, 3);
					break;
			case YahtzeeConstants.FOUR_OF_A_KIND :
					orderDice(dice);
					returnValue =  checkXOfAKind(dice, 4);
					break;
			case YahtzeeConstants.SMALL_STRAIGHT :
					straight = checkStraight(dice, 4);
					if (straight) returnValue = 30;
					break;
			case YahtzeeConstants.LARGE_STRAIGHT :
					straight = checkStraight(dice, 5);
					if (straight) returnValue = 40;
					break;
			case YahtzeeConstants.YAHTZEE :
					returnValue = checkXOfAKind(dice, 5);
					if (returnValue > 0) returnValue= 50;
					break;
			case YahtzeeConstants.FULL_HOUSE : 
					orderDice(dice);
					if (checkFullHouse(dice)) returnValue = 25;
					break;
			default : return 0;
		}
		return returnValue;
	}

	private BufferedReader openReadFile(String filename)
	{
		BufferedReader reader=null;
		try {
			reader = new BufferedReader(new FileReader(filename));
		} catch (IOException ex)
		{
			System.out.println("Cannot find that file");
		}
		return reader;
	}
	
	/* tab delimited file, score TAB persons name, top 10 scores, save the lowest score */
	private int displayTop10Scores()
	{
		/* open the file and then read from it */
		String scoresTop10;
		int lowestValue = 0, commaValue;
		BufferedReader reader = openReadFile("top10.scores");
		try {
			System.out.println("Top 10 scores :");
			while (reader.ready())
			{
				scoresTop10 = reader.readLine();
				commaValue = scoresTop10.indexOf(",");
				lowestValue = Integer.parseInt(scoresTop10.substring(0,commaValue));
				System.out.println("score : " + lowestValue +" name : "+scoresTop10.substring(commaValue+1));
			}
			reader.close();
		}catch (IOException ex)
		{
			System.out.println("There was a problem reading that file!!");
			return lowestValue;
		} catch (NumberFormatException ex)
		{
			System.out.println("The top 10 scores file, appears to be corrupted");
			return lowestValue;
		}
		return lowestValue;
	}
	
	private void updateTop10ScoresFile(int score, String username)
	{
		// could have used a separate class
		int[] scores = new int[13];
		String[] names = new String[13];
		String scoresTop10;
		int insertNumber = 0, commaNumber, scoreChecker;
		Boolean updated = false;
		BufferedReader reader = openReadFile("top10.scores");
		try {
			while (reader.ready())
			{
				scoresTop10 = reader.readLine();
				commaNumber = scoresTop10.indexOf(",");
				scoreChecker = Integer.parseInt(scoresTop10.substring(0, commaNumber));;
				System.out.println("insertNUmber "+insertNumber +" score "+ scoreChecker);
				if (scoreChecker < score && !updated)
				{
					scores[insertNumber] = score;
					names[insertNumber] = username;
					insertNumber++;
					updated = true;
				}
				scores[insertNumber] = scoreChecker;
				names[insertNumber] = scoresTop10.substring(commaNumber+1);
				insertNumber++;
			}
			reader.close();
		}catch (IOException ex)
		{
			System.out.println("There was a problem reading that file!!");
		} catch (NumberFormatException ex)
		{
			System.out.println("The top 10 scores file, appears to be corrupted");
		}
		// now we just write out the scores
		try 
		{
			PrintWriter fileWriter = new PrintWriter(new FileWriter("top10.scores"));
			for (int i =0; i < 10; i++)
				fileWriter.println(scores[i]+","+names[i]);
			fileWriter.close();
			
		} catch (IOException ex)
		{
			System.out.println("Error updating the top 10 scores file");
		}
		
	}
	
	private void playGame() {
		/* You fill this in */
		int[] dice = new int[YahtzeeConstants.N_DICE];
		int[][][] scores = new int[nPlayers][nXPlayer][N_CATEGORIES];
		int category, theScore, lowestValueTop10, boardX=0;
		IODialog dialog = getDialog();
	
		lowestValueTop10 = displayTop10Scores();
		// run through all of the categories
		setupScores(scores, nPlayers, N_CATEGORIES,nXPlayer);
		for (int catI = 0; catI < (YahtzeeConstants.N_SCORING_CATEGORIES*nXPlayer); catI++)
		{
			for (int playerI = 1; playerI <= nPlayers*nXPlayer;playerI+=nXPlayer)
			{
				setupDice(dice,true);
				display.printMessage(playerNames[playerI-1] + " turn! Click \"Roll Dice\" button to roll the dice"+(cheatMode? " !!Cheat mode!!":""));
				display.waitForPlayerToClickRoll(playerI);
				for (int playerTimes = 1; playerTimes < playerTimesNumber; playerTimes++)
				{
					display.displayDice(dice);
					display.printMessage(playerNames[playerI-1] + " pick any dice to change and then click \"Roll again\"");
					if (!cheatMode)
					{
						display.waitForPlayerToSelectDice();
						setupDice(dice,false);
					}else
						// cheating mode
					{
						for (int cheatI =0; cheatI < N_DICE; cheatI++)
							dice[cheatI] = dialog.readInt("Enter value for dice "+cheatI);
						playerTimes = playerTimesNumber;
					}
				}
				display.displayDice(dice);
				// need to order the dice, to check for categories. 
				orderDice(dice);
				// update the score for a category, if not already done.
				do {
					display.printMessage(playerNames[playerI-1] + " pick a category that you have not used before");
					category = display.waitForPlayerToSelectCategory();
					theScore = checkCategory(dice, category);
					if (nXPlayer > 1)
					{
						boardX = nXPlayer +1;
						while (boardX > nXPlayer && boardX > 0)
						{
							boardX = dialog.readInt("Please select board to place score (below "+ nXPlayer +")");
						}
					}else 
						boardX = 0;
				} while(updateScore(scores[(playerI-1)/nXPlayer],category, theScore,boardX));
				// update the displays
				display.updateScorecard(category,(playerI+boardX),scores[(playerI-1)/nXPlayer][boardX][category]);
				display.updateScorecard(TOTAL, (playerI+boardX), totalScores(scores[(playerI-1)/nXPlayer][boardX]));
			}
		}
		
		int newTotal, highestTotal=0, highestPlayer=0;
		for (int i =1; i <=nPlayers; i+=nXPlayer)
		{
			newTotal = 0;
			for (boardX = 0; boardX < nXPlayer; boardX++)
			{
				// need to update the score for the upper score > 63 + 35
				// and returns the new total for that player and also updates the lower/upper score total values
				newTotal += updateScoreAfterGame(scores[i-1][boardX]);
				if (newTotal > highestTotal) 
				{
					highestTotal = newTotal;
					highestPlayer = i;
				}
				display.updateScorecard(YahtzeeConstants.LOWER_SCORE, i+boardX, scores[i-1][boardX][YahtzeeConstants.LOWER_SCORE-1]);
				display.updateScorecard(YahtzeeConstants.UPPER_BONUS, i+boardX, scores[i-1][boardX][YahtzeeConstants.UPPER_BONUS-1]);
				display.updateScorecard(YahtzeeConstants.UPPER_SCORE, i+boardX, scores[i-1][boardX][YahtzeeConstants.UPPER_SCORE-1]);
			}
			// update the single score card
			display.updateScorecard(YahtzeeConstants.TOTAL, i, newTotal);
		}
		// display the highest player total and also congratulations..
		if (highestPlayer >0)
		{
			display.printMessage("Congratulations,"+playerNames[highestPlayer-1]+", you\'re the winner with the total score of "+highestTotal);
			if (highestTotal > lowestValueTop10)
			{
				// update the file scores if it is above the top 10 lowest value
				updateTop10ScoresFile(highestTotal, playerNames[highestPlayer-1]);
				display.printMessage("Congratulations, " + playerNames[highestPlayer-1]+", you have entered the top 10 scores");
			}
		}
	}

/* Java main method to ensure that this program starts correctly */
	public static void main(String[] args) {
		new Yahtzee().start(args);
	}

/* Private instance variables */
	private int nPlayers,nXPlayer;
	private Boolean cheatMode =false;
	private String[] playerNames;
	private YahtzeeDisplay display;
	private RandomGenerator rgen = RandomGenerator.getInstance();

}
